#include <iostream>
#include <string>
#include <thread>
#include <mutex>
#include <deque>
#include <memory>
#include <condition_variable>

using namespace std;

//implements the methods
class Servant
{
public:
  int method1(string arg) { return (arg.length() + 1); }
  int method2(string arg) { return (arg.length() + 2); }
};

//wrapper for mutex
class Mutex
{
  mutex mtx;
  volatile bool locked = false;
public:
  bool isLocked()
  {
    return locked;
  }
  void lock()
  {
    mtx.lock();
    locked = true;
  }
  void unlock()
  {
    mtx.unlock();
    locked = false;
  }
  operator mutex&() { return mtx; }
};

//allows clients to access the results of a method call on the proxy
class Future
{
private:
  int value = 0;
  volatile bool avalaible = false;
  void set(int val)
  {
    value = val;
    avalaible = true;
  }
  friend class Method1Request;
  friend class Method2Request;
public:
  bool isAvalaible() { return avalaible; }
  int get() { return value; }
};

//method request is created for each method call on the proxy
class MethodRequest
{
public:
  virtual bool guard() = 0;
  virtual void execute() = 0;
};

class Method1Request : public MethodRequest
{
private:
  Servant &servant;
  string arg;
  shared_ptr<Future> result;
  //synchronization
  static Mutex mtx;
public:
  Method1Request(Servant& srvnt, string a, shared_ptr<Future> res)
    : servant(srvnt), arg(a), result(res)
  { }
  bool guard()
  {
    return mtx.isLocked();
  }
  void execute()
  {
    mtx.lock();
    int val = servant.method1(arg);
    result->set(val);
    mtx.unlock();
  }
};

Mutex Method1Request::mtx;

class Method2Request : public MethodRequest
{
private:
  Servant &servant;
  string arg;
  shared_ptr<Future> result;
  //synchronization
  static Mutex mtx;
public:
  Method2Request(Servant& srvnt, string a, shared_ptr<Future> res)
    : servant(srvnt), arg(a), result(res)
  { }
  bool guard()
  {
    return mtx.isLocked();
  }
  void execute()
  {
    mtx.lock();
    int val = servant.method2(arg);
    result->set(val);
    mtx.unlock();
  }
};

Mutex Method2Request::mtx;

//maintains pending method requests 
class ActivationQueue
{
private:
  deque<shared_ptr<MethodRequest>> methReq;
  //synchronization
  Mutex mtx;
  condition_variable queueNotEmpty;
public:
  void insert(shared_ptr<MethodRequest> mr)
  {
    mtx.lock();
    methReq.push_back(mr);
    mtx.unlock();
    queueNotEmpty.notify_one();
  }
  shared_ptr<MethodRequest> remove()
  {
    shared_ptr<MethodRequest> ret;

    //suspends thread when methReq is empty
    unique_lock<mutex> lck(mtx);
    queueNotEmpty.wait(lck, [&]{ return methReq.size() > 0; });

    ret = methReq[0];
    methReq.pop_front();

    return ret;
  }
};

//receives method requests and executes them on the servant
class Scheduler
{
private:
  ActivationQueue act_queue;
  volatile bool finish = false;
  void stop()
  {
    finish = true;
  }
  void run()
  {
    thread([&]()
    {
      while (!finish)
      {
        shared_ptr<MethodRequest> mr = act_queue.remove();
        if (mr)
        {
          if (!mr->guard())
          {
            //execute method request in sep. thread
            thread([=]()
            {
              mr->execute();
            }).detach();
          }
          else
            act_queue.insert(mr);
        }
      }
    }).detach();
  }
public:
  Scheduler() { run(); }
  ~Scheduler() { stop(); }
  void insert(shared_ptr<MethodRequest> mr)
  {
    act_queue.insert(mr);
  }
};

//provides an interface that allows clients to access methods of an object
class Proxy
{
private:
  Scheduler scheduler;
  Servant servant;
public:
  shared_ptr<Future> method1(string arg)
  {
    shared_ptr<Future> future(new Future);
    shared_ptr<MethodRequest> mr(new Method1Request(servant, arg, future));
    scheduler.insert(mr);
    return future;
  }
  shared_ptr<Future> method2(string arg)
  {
    shared_ptr<Future> future(new Future);
    shared_ptr<MethodRequest> mr(new Method2Request(servant, arg, future));
    scheduler.insert(mr);
    return future;
  }
};

int main()
{
  static Proxy p;
  shared_ptr<Future> f1 = p.method1("aaa");
  shared_ptr<Future> f2 = p.method2("bbb");
  shared_ptr<Future> f3 = p.method2("ccc");

  while (!(f1->isAvalaible()))
    ;
  while (!(f2->isAvalaible()))
    ;
  while (!(f3->isAvalaible()))
    ;

  cout << f1->get() << endl;
  cout << f2->get() << endl;
  cout << f3->get() << endl;
}